/*
 * bkpupsd - Simple UPS daemon for APC Back-UPS Pro. series
 *
 * Copyright (C) 1999-2000, JEPRO Co. Ltd.  All right reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of JEPRO Co., Ltd. nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* bkpupsd - Simple UPS daemon for APC Back-UPS pro. series
 * Copyright (c) 1997, Yoshifumi Watanabe<mwatts@edu1.tokyo-med.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
#include <sys/cdefs.h>
#ifndef lint
__COPYRIGHT("Copyright (C) 1999-2000, JEPRO Co. Ltd.  All right reserved.");
__RCSID("$Id: bkpupsd.c,v 1.23 1999/12/25 03:46:53 taca Exp $");
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <err.h>
#include <errno.h>
#include <paths.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <syslog.h>
#include <termios.h>
#include <sys/fcntl.h>
#include <sys/wait.h>
#ifdef HAVE_UUCPLOCK
#include <libutil.h>
#else
#include "uucplock.h"
#endif
#ifdef HAVE_UPS_CONTROL
#include <sys/com_ups.h>
#endif

#define LINE_RECOVER_TIME	5

#define	FORK_RETRY		50

#ifndef LOG_BKPUPSD
#define LOG_BKPUPSD	LOG_DAEMON
#endif

#ifndef	DEFAULT_DEVICE
#define DEFAULT_DEVICE		"/dev/dty00"
#endif

#ifndef SCRIPT_DIR
#define SCRIPT_DIR		"/usr/pkg/libexec/bkpupsd"
#endif

#define _BKPUPSD_PID_PATH	"/var/run/bkpupsd.pid"
#define _SHUTDOWN_PATH		"/sbin/shutdown"
#define SHUTDOWN_WHEN		"now"
#define SHUTDOWN_HALT_ARG	"-h"
#define SHUTDOWN_POWEROFF_ARG	"-p"

static char version[] = "1.4.2";

#define UPS_SHUTDOWN		0
#define UPS_LINE_LOST		1
#define UPS_LINE_RECOVERED	2
#define UPS_BATTERY_SHUTDOWN	3

int log_local[] = {
	LOG_LOCAL0, LOG_LOCAL1, LOG_LOCAL2, LOG_LOCAL3,
	LOG_LOCAL4, LOG_LOCAL5, LOG_LOCAL6, LOG_LOCAL7,
};

char *scripts[] = {
	"line.shutdown",
	"line.lost",
	"line.recover",
	"battery.shutdown",
};

enum state {
	NORMAL, FAIL, FAIL_ALERTED, RECOVER, RECOVER_ALERTED,
};

char *ups_table[] = { "bkpro", "bk", "freqf", NULL };
#define BKPRO	0
#define	BK	1
#define	FREQF	2
#define MAX_UPS	FREQF

int ups_type = BKPRO;

int line_fail_timeout	= 0;
int line_recover_time	= LINE_RECOVER_TIME;
int make_power_off	= 0;
int ups_sleep_on_exit	= 0;
int dont_sleep_ups	= 0;
int ups_fd		= -1;
#ifdef HAVE_UPS_CONTROL
int kernel_mode		= 0;
#endif

char *uulock = NULL;

int main __P((int argc, char **argv));
void bkpupsd __P((int fd, int nolow, int testmode));
void usage __P((int status));
static void execomand __P((int fd, int cmd, int testing));
static int open_tty __P((char *name, int mode));
static void makepid __P((void));
static void waitchildren __P((int sig));
static void quit __P((int sig));
static int mysleep __P((unsigned int second));
static int tty_get __P((int fd, unsigned long request, int *value));
static int tty_set __P((int fd, unsigned long request, int value));

static int ups_init __P((int fd));
static int ups_status __P((int fd));
static int ups_sleep __P((int fd));

int main(argc, argv)
int argc;
char  **argv;
{
	int ch;
	int fd;
	int n;
	int log_fac;
	int test_mode, no_low_mode, debug;
	char *ups_device = DEFAULT_DEVICE;

	test_mode = no_low_mode = 0;
	log_fac = LOG_BKPUPSD;
	debug = 0;
	while ((ch = getopt(argc, argv, "cdf:hklnopr:s:t:u:v")) != -1) {
		switch (ch) {
		case 'c':
			ups_type = BK;
			break;
		case 'd':
			debug = 1;
			break;
		case 'f':
			ups_device = optarg;
			break;
		case 'h':
			usage(EXIT_SUCCESS);
			break;
#ifdef HAVE_UPS_CONTROL
		case 'k':
			kernel_mode = 1;
			break;
#endif
		case 'l':
			no_low_mode = 1;
			break;
		case 'n':
			test_mode = 1;
			break;
		case 'o':
			make_power_off = 1;
			break;
		case 'p':
			dont_sleep_ups = 1;
			break;
		case 'r':
			n = atoi(optarg);
			if (n > 0)
				line_recover_time = n;
			break;
		case 's':
			n = atoi(optarg);
			if (n >= 0 && n <= 7)
				log_fac = log_local[n];
			break;
		case 't':
			n = atoi(optarg);
			if (n > 0)
				line_fail_timeout = n;
			break;
		case 'u':
			for (n = 0; ups_table[n] != NULL; n++) {
				if (strcasecmp(ups_table[n], optarg) == 0)
					break;
			}
			if (n > MAX_UPS) {
				errx(EXIT_FAILURE,
				     "unknown UPS type \"%s\"\n", optarg);
			}
			ups_type = n;
			break;
		case 'v':
			printf("bkpupsd version %s\n", version);
			exit(0);
			break;
		default:
			warnx("don't know -%c option\n", ch);
			usage(EXIT_FAILURE);
		}
	}
	if (no_low_mode && line_fail_timeout == 0) {
		errx(EXIT_FAILURE,
		     "Please specify timeout parameter with -t option.");
	}

	argc -= optind;
	argv += optind;
	if (argc == 2) {
		ups_device = argv[1];
	}

	/* Detach and Become to a daemon process */
	if (!debug && daemon(0, 0) < 0) {
		err(1, "daemon");
		exit(1);
	}

	openlog("bkpupsd", LOG_PID, log_fac);

	fd = open_tty(ups_device, O_RDWR | O_NDELAY);
	if (fd < 0) {
		syslog(LOG_ERR, "%s: %m", ups_device);
		quit(1);
	}

	makepid();

	signal(SIGHUP, quit);
	signal(SIGINT, quit);
	signal(SIGTERM, quit);

	/* Initialize UPS */
	if (ups_init(fd) < 0) {
		syslog(LOG_ERR, "failed to initializing UPS.");
		quit(1);
	}

	/* wait for children */
	signal(SIGCHLD, waitchildren);

	bkpupsd(fd, no_low_mode, test_mode);
	return 0;
}

void
bkpupsd(fd, no_low, test_mode)
	int fd;
	int no_low;
	int test_mode;
{
	enum state line_state, battery_state;
	int status = 0;
	time_t line_failed = 0, power_recovered = 0, battery_failed = 0;
	time_t now;
	int power_fail, battery_fail;

	syslog(LOG_NOTICE, "bkpupsd %s started for %s UPS", version,
	       ups_table[ups_type]);
	if (no_low)
		syslog(LOG_NOTICE, "without check battery low.");
	if (test_mode)
		syslog(LOG_NOTICE, "in test mode.");

	power_fail = battery_fail = 0;
	line_state = battery_state = NORMAL;

	while (1) {
		status = mysleep(1);
		if (status > 1)
			continue;
		else if (status < 0) {
			syslog(LOG_ERR, "select: %m");
			quit(1);
		}
			
		status = ups_status(fd);
		time(&now);

		if (ups_type == BK) {
			power_fail = (status & TIOCM_CTS)? 1: 0;
			battery_fail = (status & TIOCM_CAR)? 1: 0;
		} else if (ups_type == BKPRO) {
			power_fail = (status & TIOCM_RNG)? 1: 0;
			battery_fail = (status & TIOCM_CAR)? 1: 0;
		} else if (ups_type == FREQF) {
			power_fail = (status & TIOCM_CTS)? 0: 1;
			battery_fail = (status & TIOCM_CAR)? 0: 1;
		}

		switch (line_state) {
		case NORMAL:
			if (power_fail) {
				line_failed = now;
				line_state = FAIL;
			}
			break;
		case FAIL:
			if (power_fail) {
				syslog(LOG_ALERT, "line power has lost.");
				execomand(fd, UPS_LINE_LOST, test_mode);
				line_state = FAIL_ALERTED;
			} else {
				line_failed = 0;
				line_state = NORMAL;
			}
			break;
		case FAIL_ALERTED:
			if (power_fail) {
				if (line_fail_timeout &&
				    now > line_failed + line_fail_timeout) {
					/* power fail timeout */
					syslog(LOG_EMERG, "shutdown for line power failure timeout.");
					if (!test_mode) {
						ups_sleep_on_exit = 1;
						ups_fd = fd;
					}
					execomand(fd, UPS_SHUTDOWN, test_mode);
					quit(0);
				}
			} else {
				power_recovered = now;
				line_state = RECOVER;
			}
			break;
		case RECOVER:
			if (power_fail) {
				line_state = FAIL_ALERTED;
			} else {
				if (now > power_recovered + line_recover_time) {
					syslog(LOG_ALERT, "line power has recovered.");
					execomand(fd, UPS_LINE_RECOVERED,
						  test_mode);
					line_state = NORMAL;
				}
			}
			break;
		default:
			syslog(LOG_ERR, "panic: no such power state %d",
			       line_state);
			quit(1);
		}

		if (no_low || line_state != FAIL_ALERTED) {
			battery_fail = NORMAL;
			continue;
		}

		switch (battery_state) {
		case NORMAL:
			if (battery_fail) {
				battery_failed = now;
				battery_state = FAIL;
			}
			break;
		case FAIL:
			if (battery_fail) {
				syslog(LOG_ALERT, "battery is low.");
				if (!test_mode) {
					ups_sleep_on_exit = 1;
					ups_fd = fd;
				}
				execomand(fd, UPS_BATTERY_SHUTDOWN, test_mode);
				quit(0);
			} else {
				battery_failed = 0;
				battery_state = NORMAL;
			}
			break;
		default:
			syslog(LOG_ERR, "panic: no such battery state %d",
			       battery_state);
			quit(1);
		}
	}
	quit(2);		/* should not come to here. */
}

void
usage(status)
	int status;
{
	fputs("Usage: bkpupsd [-cdhlnv] [-f device] [-r seconds] [-s number] [-t seconds]\n", stderr);
	if (status == EXIT_SUCCESS) {
		fputs("\t-c: Select cable type.\n", stderr);
		fputs("\t-d: Debug mode, don't detach from terminal.\n",
		      stderr);
		fputs("\t-f: Specify serial device.\n", stderr);
		fputs("\t-h: Show help, this message.\n", stderr);
		fputs("\t-l: Ignore completely low battery status.\n", stderr);
		fputs("\t-n: Don't execute external script for testing only.\n",
		      stderr);
		fputs("\t-p: Never make ups in sleep mode.\n", stderr);
		fputs("\t-r: Specify the line recovery time.\n", stderr);
		fputs("\t-s: Specify syslog facility, LOCAL_#.\n", stderr);
		fputs("\t-t: Specify UPS shutdown time from line has lost.\n",
		      stderr);
		fputs("\t-t: Specify UPS type (bkpro, bk, freqf).\n", stderr);
		fputs("\t-v: Show version.\n", stderr);
	}
	exit(status);
}

/* Subroutine to execute shell script command */

static void
execomand(fd, cmd, test_mode)
	int fd;
	int cmd;
	int test_mode;
{
	int pid, n;
	struct stat st;
	char command[BUFSIZ];

	if (cmd < 0 || cmd >= (sizeof scripts / sizeof *scripts)) {
		syslog(LOG_ERR, "illegal cmd %d", cmd);
		return;
	}
	sprintf(command, "%s/%s", SCRIPT_DIR, scripts[cmd]);

	if (test_mode) {
		if (stat(command, &st) < 0) {
			syslog(LOG_ERR, "%s: %m", command);
			return;
		}
		syslog(LOG_NOTICE, "execute %s (testing)", command);
		return;
	}

	n = 0;
	while ((pid = fork()) < 0) {
		syslog(LOG_ERR, "fork failed: %m");
		n++;
		if (errno != EAGAIN && errno != ENOMEM && n > FORK_RETRY) {
			syslog(LOG_EMERG,
			       "bkpupsd exit without executing the command.");
			quit(1);
		}
		mysleep(1);
	}

	if (pid == 0) {
		close(fd);
		execl(_PATH_BSHELL, "sh", "-c", command, NULL);
		syslog(LOG_ERR, "execution failed: %m");
		exit(1);
	}
	/* parent */
	syslog(LOG_NOTICE, "execute %s", command);
}


static int
open_tty(name, mode)
	char *name;
	int mode;
{
	int fd, r;
	char *s;

	s = strrchr(name, '/');
	if (s != NULL)
		s++;
	else
		s = name;
	r = uu_lock(s);
	if (r < 0) {
		syslog(LOG_ERR, "bkpupsd is already running");
		quit(1);
	}
	uulock = s;
	fd = open(name, mode);
	return fd;
}

/* Subroutine to make bkpupsd.pid to /var/run/ directory. */
static void
makepid()
{
	char *path = _BKPUPSD_PID_PATH;
	int fd;
	char buf[BUFSIZ];

	fd = open(path, O_RDWR | O_CREAT | O_TRUNC);
	if (fd < 0) {
		syslog(LOG_ERR, "%s: open: %m", path);
		exit(1);
	}
	sprintf(buf, "%d\n", getpid());
	if (write(fd, buf, strlen(buf)) < 0) {
		syslog(LOG_ERR, "%s: write: %m", path);
		quit(1);
	}
	close(fd);
}

/*
 * wait handler.
 */
static void
waitchildren(sig)
	int sig;
{
	int status;

	while (wait3(&status, WNOHANG, NULL) == 0 && WIFEXITED(status))
	       ;
}

/*
 * signal handler.
 */
static void
quit(sig)
	int sig;
{
	(void)uu_unlock(uulock);
	unlink(_BKPUPSD_PID_PATH);
	if (ups_sleep_on_exit && ups_fd >= 0) {
		signal(SIGHUP, SIG_IGN);
		signal(SIGINT, SIG_IGN);
		signal(SIGTERM, SIG_IGN);
		mysleep(3);
		syslog(LOG_DEBUG, "make a ups sleep.");
		sync(); sync();
		if (!dont_sleep_ups)
			ups_sleep(ups_fd);
		syslog(LOG_NOTICE, "bkpupsd halt the system.");
		sync(); sync();
		if (make_power_off)
			execl(_SHUTDOWN_PATH, _SHUTDOWN_PATH,
			      SHUTDOWN_HALT_ARG, SHUTDOWN_POWEROFF_ARG,
			      SHUTDOWN_WHEN, NULL);
		else
			execl(_SHUTDOWN_PATH, _SHUTDOWN_PATH,
			      SHUTDOWN_HALT_ARG, SHUTDOWN_WHEN, NULL);
		syslog(LOG_ERR, "%s: %m", _SHUTDOWN_PATH);
		_exit(EXIT_FAILURE);
	}
	syslog(LOG_NOTICE, "bkpupsd exit.");
	exit(sig);
}

static int
tty_get(fd, request, value)
	int fd;
	unsigned long request;
	int *value;
{
	int s;

	s = ioctl(fd, request, value);
	if (s < 0) {
		syslog(LOG_DEBUG, "ioctl get (%x, %o): %m", request, value);
		return -1;
	}
	return 0;
}

static int
tty_set(fd, request, value)
	int fd;
	unsigned long request;
	const int value;
{
	int s;

	s = ioctl(fd, request, &value);
	if (s < 0) {
		syslog(LOG_DEBUG, "ioctl set (%x, %o): %m", request, value);
		return -1;
	}
	return 0;
}

/*
 * Initialize UPS
 */
static int
ups_init(fd)
	int fd;
{
	if (ups_type == BK) {
		if (tty_set(fd, TIOCMBIC, TIOCM_DTR) < 0)
			return -1;
	} else if (ups_type == BKPRO) {
		if (tty_set(fd, TIOCMBIC, TIOCM_RTS) < 0 ||
		    tty_set(fd, TIOCMBIS, TIOCM_DTR) < 0)
			return -1;
	} else if (ups_type == FREQF) {
		if (tty_set(fd, TIOCMBIC, TIOCM_DTR) < 0 ||
		    tty_set(fd, TIOCMBIS, TIOCM_RTS) < 0)
			return -1;
	}
	return 0;
}

/*
 * Get UPS status
 */
static int
ups_status(fd)
	int fd;
{
	int status;
	static int ostatus = -1;

	if (ups_type == BKPRO) {
		tty_set(fd, TIOCMBIC, TIOCM_RTS);
		tty_set(fd, TIOCMBIC, TIOCM_DTR);
		tty_set(fd, TIOCMBIS, TIOCM_DTR);
	}

	tty_get(fd, TIOCMGET, &status);

	if (status != ostatus) {
		if (ostatus == -1)
			syslog(LOG_DEBUG, "serial status now %o", status);
		else
			syslog(LOG_DEBUG, "serial status now %o => %o",
			       ostatus, status);
		ostatus = status;
	}

	return status;
}

/*
 * Set UPS in sleep mode
 */
static int
ups_sleep(fd)
	int fd;
{
	int mode, setbits, resetbits;

	setbits = resetbits = 0;
	if (ups_type == BK) {
		setbits = TIOCM_DTR;
	} else if (ups_type == BKPRO) {
		resetbits = TIOCM_DTR;
		setbits = TIOCM_DTR;
		setbits = TIOCM_RTS;
	} else if (ups_type == FREQF) {
		setbits = TIOCM_DTR;
	} else {
		errx(EXIT_FAILURE, "unknown UPS type \"%d\"\n", ups_type);
	}

#ifdef HAVE_UPS_CONTROL
	if (kernel_mode) {
		if (tty_get(fd, TIOCMGET, &mode) < 0)
			return -1;
		mode &= ~resetbits;
		mode |= setbits;
		if (ioctl(fd, UPSIOC_SET, &mode) < 0) {
			syslog(LOG_DEBUG, "UPSIOC_SET(%o): %m", mode);
			return -1;
		}
	} else
#endif
	{
		tty_set(fd, TIOCMBIC, resetbits);
		tty_set(fd, TIOCMBIS, setbits);
	}
	tty_get(fd, TIOCMGET, &mode);
	syslog(LOG_DEBUG, "serial status now %o", mode);
	return 0;
}

static int
mysleep(second)
	unsigned int second;
{
	struct timeval timeout;

	timeout.tv_sec = 1;
	timeout.tv_usec = 0;

	if (select(0, NULL, NULL, NULL, &timeout) < 0) {
		if (errno == EINTR)
			return 1;
	}
	return 0;
}
